from src.matching.matching_base import MatchingBase
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

class TfidfCosineMatcher(MatchingBase):
    def computer_simlarity(self, query_info, retrived_indices):
        vectorizer = TfidfVectorizer()
        #TODO should fit on all corpus
        query = query_info.terminfo
        samples = [self.dict_map['corpus'][2][idx] for idx in retrived_indices]
        docs_tfidf = vectorizer.fit_transform(samples)
        query_tfidf = vectorizer.transform(query)
        return self._calc_cos(docs_tfidf, query_tfidf)

    def _calc_cos(self, docs, query):
        cosineSimilarities = cosine_similarity(query, docs).flatten()
        return cosineSimilarities




from src.matching.matching_base import MatchingBase
from rank_bm25 import BM25Okapi

class BM25Matcher(MatchingBase):

    def compute_simlarity(self, query_info, retrived_indices):
        #TODO should find a better bm25 package
        '''query and samples should be cleaned and tokenized'''
        query = query_info.terminfo
        samples = [self.dict_map['corpus'][2][idx] for idx in retrived_indices]
        self.bm25 = BM25Okapi(samples)
        scores = self.bm25.get_scores(query)
        return scores





import faiss
import numpy as np
from src.retrieval.retrieval_base import RetrievalBase

class FaissIndex(RetrievalBase):
    def __init__(self, config):
        if 'vec_size' not in config:
            raise KeyError("RetrievalFaiss plugin not have vec_size attribute")
        self.vec_size = config['vec_size']
        self.top_k = 50
        if 'top_k' in config:
            self.top_k = int(config['top_k'])
        index = faiss.IndexFlatL2(self.vec_size)
        self.indexidmap = faiss.IndexIDMap(index)

    def _build_index(self, corpus):
        kvdict = {idx: items[0] for idx, items in corpus[3].items()}
        for idx, vec in  kvdict.items():
            self.indexidmap.add_with_ids(vec.reshape(1, -1), np.array([idx]))
        print(f"Faiss_Index: count ntotal:{self.indexidmap.ntotal}")

    def add_index(self, corpus):
        ntotal = self.indexidmap.ntotal
        kvdict = {idx+ntotal: items[0] for idx, items in corpus[3].items()}
        for idx, vec in kvdict.items():
            self.indexidmap.add_with_ids(vec.reshape(1, -1), np.array([idx]))
        print("add {} new data, current faiss total {}".format(len(corpus), self.indexidmap.ntotal))

    def search(self, queryinfo, resultinfo):
        # 搜索与当前向量最相近的k个向量的索引
        _ , ids = self.indexidmap.search(queryinfo.embinfo[0].reshape(1, -1), self.top_k)
        resultinfo.retrieval['faiss'] = ids[0]
        resultinfo.retrieval['allids'] = ids[0]



from annoy import AnnoyIndex
from src.retrieval.retireval_base import RetrievalBase

class AnnoyIndex(RetrievalBase):
    def __init__(self, config):
        if 'vec_size' not in config:
            raise KeyError("RetrievalAnnoy plugin not have vec_size attribute")
        self.vec_size = config['vec_size']
        self.top_k = 50
        if 'top_k' in config:
            self.top_k = config['top_k']
        self.n_trees = 15
        if 'n_trees' in config:
            self.n_trees = config['n_trees']
        self.annoyindex = AnnoyIndex(self.vec_size)

    def _build_index(self, corpus):
        pass

    def search(self, queryinfo, resultinfo):
        pass



from src.matching.matching_base import MatchingBase

class WordJaccardMatcher(MatchingBase):
    def compute_simlarity(self, query_info, retrived_indices):
        seg_query = query_info.terminfo
        seg_samples = [self.dict_map['corpus'][2][sample] for sample in retrived_indices]
        return [word_jaccard(seg_query, sample) for sample in seg_samples]

class CharJaccardMatcher(MatchingBase):
    def compute_simlarity(self, query_info, retrived_indices):
        query = query_info.cleantext
        samples = [self.dict_map['corpus'][0][sample] for sample in retrived_indices]
        return [char_jaccard(query, sample) for sample in samples]

def word_jaccard(seg1, seg2):
    a = list(set(seg1).intersection(set(seg2)))
    b = list(set(seg1).union(set(seg2)))
    return float(len(a) / len(b))


def char_jaccard(sen1, sen2):
    a = list(set(list(sen1)).intersection(set(list(sen2))))
    b = list(set(list(sen1)).union(set(list(sen2))))
    return float(len(a) / len(b))




class AnalysisBase:
    def process(self, text):
        raise NotImplementedError(
            "Analysis plugin must have sigle-process function")

    def value_process(self, data_dict):
        '''
        data_dict repr {idx1: text1,  idx2:text2...}
        '''
        return {idx: self.process(text) for idx, text in data_dict.items()}



# -*- coding:utf-8 -*-


import os
import re
import string
from src.analysis.analysis_base import AnalysisBase


class BathroomSSC(AnalysisBase):
    """
        Description:
            preprocess str, drop punction and prefix-suffix
    """

    def __init__(self, dict_map, config):
        self.exclude_set = []
        self.prefix_list = []
        self.suffix_list = []
        self.prefix_pattern = None
        if 'punction' in config['use_dict_name']:
            self.exclude_set = dict_map['punction']
        if 'prefix' in config['use_dict_name']:
            self.prefix_list = dict_map['prefix']
            self.prefix_pattern = re.compile(
                r"^{}".format("|".join(self.prefix_list)))
        if 'suffix' in config['use_dict_name']:
            self.suffix_list = config['use_dict_name']

    def drop_endpunc(self, text):
        while text:
            if text[-1] in self.exclude_set:
                text = text[:-1]
            else:
                break
        return text

    def remove_suffix(self, text):
        if isinstance(text, str):
            if len(text) > 0:
                start = len(text)
                if text[-1] in self.suffix_list:
                    text = text[:-1]
                if start > len(text):
                    return self.remove_suffix(text)
                else:
                    return text
            else:
                return text
        else:
            return text

    def clean(self, text):
        text_clean = text.strip()
        if self.exclude_set:
            text_clean = self.drop_endpunc(text_clean.lower().replace("\n", ""))
        if self.prefix_pattern:
            text_clean = re.sub(self.prefix_pattern, '', text_clean)
        if self. suffix_list:
            text_clean = self.remove_suffix(text_clean)
        return text_clean

    def process(self, text):
        return self.clean(text)



# -*- coding:utf-8 -*-

import os
import logging
from logging import FileHandler
import time

class SafeFileHandler(FileHandler):
    """solve multi-process write file problems"""
    def __init__(self, filename, mode, encoding=None, delay=0):
        """
        Use the specified filename for streamed logging
        """
        FileHandler.__init__(self, filename, mode, encoding, delay)
        self.mode = mode
        self.encoding = encoding
        self.suffix = "%Y%m%d"
        self.suffix_time = ""

    def emit(self, record):
        """
        Emit a record.

        Always check time
        """
        try:
            if self.check_baseFilename(record):
                self.build_baseFilename()
            FileHandler.emit(self, record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def check_baseFilename(self, record):
        """
        Determine if builder should occur.

        record is not used, as we are just comparing times,
        but it is needed so the method signatures are the same
        """
        timeTuple = time.localtime()

        if self.suffix_time != time.strftime(self.suffix, timeTuple) or not os.path.exists(self.baseFilename+'.'+self.suffix_time):
            return 1
        else:
            return 0

    def build_baseFilename(self):
        """
        do builder; in this case,
        old time stamp is removed from filename and
        a new time stamp is append to the filename
        """
        if self.stream:
            self.stream.close()
            self.stream = None

        # remove old suffix
        if self.suffix_time != "":
            index = self.baseFilename.find("."+self.suffix_time)
            if index == -1:
                index = self.baseFilename.rfind(".")
            self.baseFilename = self.baseFilename[:index]

        # add new suffix
        currentTimeTuple = time.localtime()
        self.suffix_time = time.strftime(self.suffix, currentTimeTuple)
        self.baseFilename  = self.baseFilename + "." + self.suffix_time

        self.mode = 'a'
        if not self.delay:
            self.stream = self._open()


def setlogger(config):
    cfg = config
    logger = logging.getLogger("QAsearch")
    formatter = logging.Formatter(cfg['LogFormat'])
    logger.setLevel(10)
    # setup console setting
    if cfg['LogLevel_Console'] != 0:
        ch = logging.StreamHandler()
        ch.setLevel(cfg["LogLevel_Console"])
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    # setup file logging path
    if cfg['LogLevel_File'] != 0:
        folder = cfg['LogFile']
        if not os.path.exists(os.path.dirname(folder)):
            os.makedirs(os.path.dirname(folder), exist_ok=True)
        fh = SafeFileHandler(folder, mode='a')
        fh.setLevel(cfg["LogLevel_File"])
        fh.setFormatter(formatter)
        logger.addHandler(fh)

logger = logging.getLogger("QAsearch")

if __name__ == '__main__':
    config = {}
    logger = setlogger(config)
    text = input("input print info:")
    logger.info(text)
    logger.warning("it worked")



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BERT finetuning runner."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import time
import collections
import csv
import os
import src.matching.semantic.bert.tokenization as tokenization
import tensorflow as tf

class InputExample(object):
  """A single training/test example for simple sequence classification."""

  def __init__(self, guid, text_a, text_b=None, label=None):
    """Constructs a InputExample.
    Args:
      guid: Unique id for the example.
      text_a: string. The untokenized text of the first sequence. For single
        sequence tasks, only this sequence must be specified.
      text_b: (Optional) string. The untokenized text of the second sequence.
        Only must be specified for sequence pair tasks.
      label: (Optional) string. The label of the example. This should be
        specified for train and dev examples, but not for test examples.
    """
    self.guid = guid
    self.text_a = text_a
    self.text_b = text_b
    self.label = label


class PaddingInputExample(object):
  """Fake example so the num input examples is a multiple of the batch size.
  When running eval/predict on the TPU, we need to pad the number of examples
  to be a multiple of the batch size, because the TPU requires a fixed batch
  size. The alternative is to drop the last batch, which is bad because it means
  the entire output data won't be generated.
  We use this class instead of `None` because treating `None` as padding
  battches could cause silent errors.
  """


class InputFeatures(object):
  """A single set of features of data."""

  def __init__(self,
               input_ids,
               input_mask,
               segment_ids,
               label_id,
               is_real_example=True):
    self.input_ids = input_ids
    self.input_mask = input_mask
    self.segment_ids = segment_ids
    self.label_id = label_id
    self.is_real_example = is_real_example


class DataProcessor(object):
  """Base class for data converters for sequence classification data sets."""

  def get_train_examples(self, data_dir):
    """Gets a collection of `InputExample`s for the train set."""
    raise NotImplementedError()

  def get_dev_examples(self, data_dir):
    """Gets a collection of `InputExample`s for the dev set."""
    raise NotImplementedError()

  def get_test_examples(self, data_dir):
    """Gets a collection of `InputExample`s for prediction."""
    raise NotImplementedError()

  def get_labels(self):
    """Gets the list of labels for this data set."""
    raise NotImplementedError()

  @classmethod
  def _read_tsv(cls, input_file, quotechar=None):
    """Reads a tab separated value file."""
    with tf.gfile.Open(input_file, "r") as f:
      reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
      lines = []
      for line in reader:
        lines.append(line)
      return lines

def convert_single_example(ex_index, example, label_list, max_seq_length,
                           tokenizer):
  """Converts a single `InputExample` into a single `InputFeatures`."""

  if isinstance(example, PaddingInputExample):
    return InputFeatures(
        input_ids=[0] * max_seq_length,
        input_mask=[0] * max_seq_length,
        segment_ids=[0] * max_seq_length,
        label_id=0,
        is_real_example=False)

  label_map = {}
  for (i, label) in enumerate(label_list):
    label_map[label] = i

  tokens_a = tokenizer.tokenize(example.text_a)
  tokens_b = None
  if example.text_b:
    tokens_b = tokenizer.tokenize(example.text_b)

  if tokens_b:
    # Modifies `tokens_a` and `tokens_b` in place so that the total
    # length is less than the specified length.
    # Account for [CLS], [SEP], [SEP] with "- 3"
    _truncate_seq_pair(tokens_a, tokens_b, max_seq_length - 3)
  else:
    # Account for [CLS] and [SEP] with "- 2"
    if len(tokens_a) > max_seq_length - 2:
      tokens_a = tokens_a[0:(max_seq_length - 2)]

  # The convention in BERT is:
  # (a) For sequence pairs:
  #  tokens:   [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
  #  type_ids: 0     0  0    0    0     0       0 0     1  1  1  1   1 1
  # (b) For single sequences:
  #  tokens:   [CLS] the dog is hairy . [SEP]
  #  type_ids: 0     0   0   0  0     0 0
  #
  # Where "type_ids" are used to indicate whether this is the first
  # sequence or the second sequence. The embedding vectors for `type=0` and
  # `type=1` were learned during pre-training and are added to the wordpiece
  # embedding vector (and position vector). This is not *strictly* necessary
  # since the [SEP] token unambiguously separates the sequences, but it makes
  # it easier for the model to learn the concept of sequences.
  #
  # For classification tasks, the first vector (corresponding to [CLS]) is
  # used as the "sentence vector". Note that this only makes sense because
  # the entire model is fine-tuned.
  tokens = []
  segment_ids = []
  tokens.append("[CLS]")
  segment_ids.append(0)
  for token in tokens_a:
    tokens.append(token)
    segment_ids.append(0)
  tokens.append("[SEP]")
  segment_ids.append(0)

  if tokens_b:
    for token in tokens_b:
      tokens.append(token)
      segment_ids.append(1)
    tokens.append("[SEP]")
    segment_ids.append(1)

  input_ids = tokenizer.convert_tokens_to_ids(tokens)

  # The mask has 1 for real tokens and 0 for padding tokens. Only real
  # tokens are attended to.
  input_mask = [1] * len(input_ids)

  # Zero-pad up to the sequence length.
  while len(input_ids) < max_seq_length:
    input_ids.append(0)
    input_mask.append(0)
    segment_ids.append(0)

  assert len(input_ids) == max_seq_length
  assert len(input_mask) == max_seq_length
  assert len(segment_ids) == max_seq_length

  label_id = label_map[example.label]
  if ex_index < 5:
    tf.logging.info("*** Example ***")
    tf.logging.info("guid: %s" % (example.guid))
    tf.logging.info("tokens: %s" % " ".join(
        [tokenization.printable_text(x) for x in tokens]))
    tf.logging.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
    tf.logging.info("input_mask: %s" % " ".join([str(x) for x in input_mask]))
    tf.logging.info("segment_ids: %s" % " ".join([str(x) for x in segment_ids]))
    tf.logging.info("label: %s (id = %d)" % (example.label, label_id))

  feature = InputFeatures(
      input_ids=input_ids,
      input_mask=input_mask,
      segment_ids=segment_ids,
      label_id=label_id,
      is_real_example=True)
  return feature


def file_based_convert_examples_to_features(
    examples, label_list, max_seq_length, tokenizer, output_file):
  """Convert a set of `InputExample`s to a TFRecord file."""

  writer = tf.python_io.TFRecordWriter(output_file)
  cumsum_time = 0
  for (ex_index, example) in enumerate(examples):
    if ex_index % 10000 == 0:
      tf.logging.info("Writing example %d of %d" % (ex_index, len(examples)))
    start_time = time.time()
    feature = convert_single_example(ex_index, example, label_list,
                                     max_seq_length, tokenizer)

    def create_int_feature(values):
      f = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
      return f

    features = collections.OrderedDict()
    features["input_ids"] = create_int_feature(feature.input_ids)
    features["input_mask"] = create_int_feature(feature.input_mask)
    features["segment_ids"] = create_int_feature(feature.segment_ids)
    features["label_ids"] = create_int_feature([feature.label_id])
    features["is_real_example"] = create_int_feature(
        [int(feature.is_real_example)])
    end_time = time.time()
    wait_time = end_time - start_time
    cumsum_time += wait_time
    tf_example = tf.train.Example(features=tf.train.Features(feature=features))
    writer.write(tf_example.SerializeToString())
  writer.close()
  tf.logging.info("convert cost time: %f", cumsum_time)

def file_based_input_fn_builder(input_file, seq_length, is_training,
                                drop_remainder):
  """Creates an `input_fn` closure to be passed to TPUEstimator."""

  name_to_features = {
      "input_ids": tf.FixedLenFeature([seq_length], tf.int64),
      "input_mask": tf.FixedLenFeature([seq_length], tf.int64),
      "segment_ids": tf.FixedLenFeature([seq_length], tf.int64),
      "label_ids": tf.FixedLenFeature([], tf.int64),
      "is_real_example": tf.FixedLenFeature([], tf.int64),
  }

  def _decode_record(record, name_to_features):
    """Decodes a record to a TensorFlow example."""
    example = tf.parse_single_example(record, name_to_features)

    # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
    # So cast all int64 to int32.
    for name in list(example.keys()):
      t = example[name]
      if t.dtype == tf.int64:
        t = tf.to_int32(t)
      example[name] = t

    return example

  def input_fn(params):
    """The actual input function."""
    batch_size = params["batch_size"]

    # For training, we want a lot of parallel reading and shuffling.
    # For eval, we want no shuffling and parallel reading doesn't matter.
    d = tf.data.TFRecordDataset(input_file)
    if is_training:
      d = d.repeat()
      d = d.shuffle(buffer_size=100)

    d = d.apply(
        tf.contrib.data.map_and_batch(
            lambda record: _decode_record(record, name_to_features),
            batch_size=batch_size,
            drop_remainder=drop_remainder))

    return d

  return input_fn


def _truncate_seq_pair(tokens_a, tokens_b, max_length):
  """Truncates a sequence pair in place to the maximum length."""

  # This is a simple heuristic which will always truncate the longer sequence
  # one token at a time. This makes more sense than truncating an equal percent
  # of tokens from each, since if one sequence is very short then each token
  # that's truncated likely contains more information than a longer sequence.
  while True:
    total_length = len(tokens_a) + len(tokens_b)
    if total_length <= max_length:
      break
    if len(tokens_a) > len(tokens_b):
      tokens_a.pop()
    else:
      tokens_b.pop()

# This function is not used by this file but is still used by the Colab and
# people who depend on it.
def input_fn_builder(features, seq_length, is_training, drop_remainder):
  """Creates an `input_fn` closure to be passed to TPUEstimator."""

  all_input_ids = []
  all_input_mask = []
  all_segment_ids = []
  all_label_ids = []

  for feature in features:
    all_input_ids.append(feature.input_ids)
    all_input_mask.append(feature.input_mask)
    all_segment_ids.append(feature.segment_ids)
    all_label_ids.append(feature.label_id)

  def input_fn(params):
    """The actual input function."""
    batch_size = params["batch_size"]

    num_examples = len(features)

    # This is for demo purposes and does NOT scale to large data sets. We do
    # not use Dataset.from_generator() because that uses tf.py_func which is
    # not TPU compatible. The right way to load data is with TFRecordReader.
    d = tf.data.Dataset.from_tensor_slices({
        "input_ids":
            tf.constant(
                all_input_ids, shape=[num_examples, seq_length],
                dtype=tf.int32),
        "input_mask":
            tf.constant(
                all_input_mask,
                shape=[num_examples, seq_length],
                dtype=tf.int32),
        "segment_ids":
            tf.constant(
                all_segment_ids,
                shape=[num_examples, seq_length],
                dtype=tf.int32),
        "label_ids":
            tf.constant(all_label_ids, shape=[num_examples], dtype=tf.int32),
    })

    if is_training:
      d = d.repeat()
      d = d.shuffle(buffer_size=100)

    d = d.batch(batch_size=batch_size, drop_remainder=drop_remainder)
    return d

  return input_fn

class LCQMCPairClassificationProcessor(DataProcessor): # TODO NEED CHANGE2
  """Processor for the internal data set. sentence pair classification"""
  def __init__(self):
    self.language = "zh"

  def get_train_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "train.txt")), "train")
    # dev_0827.tsv

  def get_dev_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "dev.txt")), "dev")

  def get_test_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "test.txt")), "test")

  def get_labels(self):
    """See base class."""
    return ["0", "1"]
    #return ["-1","0", "1"]

  def _create_examples(self, lines, set_type):
    """Creates examples for the training and dev sets."""
    examples = []
    print("length of lines:",len(lines))
    cumsum_time = 0
    for (i, line) in enumerate(lines):
      #print('#i:',i,line)
      if i == 0:
        continue
      guid = "%s-%s" % (set_type, i)
      start_time = time.time()
      try:
          label = tokenization.convert_to_unicode(line[2])
          text_a = tokenization.convert_to_unicode(line[0])
          text_b = tokenization.convert_to_unicode(line[1])
          examples.append(
              InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
      except Exception:
          print('###error.i:', i, line)
      end_time = time.time()
      wait_time = end_time-start_time
      cumsum_time += wait_time
    return examples

class SentencePairClassificationProcessor(DataProcessor):
  """Processor for the internal data set. sentence pair classification"""
  def __init__(self):
    self.language = "zh"

  def get_train_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "train_0827.tsv")), "train")
    # dev_0827.tsv

  def get_dev_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "dev_0827.tsv")), "dev")

  def get_test_examples(self, data_dir):
    """See base class."""
    return self._create_examples(
        self._read_tsv(os.path.join(data_dir, "test_0827.tsv")), "test")

  def get_labels(self):
    """See base class."""
    return ["0", "1"]
    #return ["-1","0", "1"]

  def _create_examples(self, lines, set_type):
    """Creates examples for the training and dev sets."""
    examples = []
    print("length of lines:",len(lines))
    for (i, line) in enumerate(lines):
      #print('#i:',i,line)
      if i == 0:
        continue
      guid = "%s-%s" % (set_type, i)
      try:
          label = tokenization.convert_to_unicode(line[0])
          text_a = tokenization.convert_to_unicode(line[1])
          text_b = tokenization.convert_to_unicode(line[2])
          examples.append(
              InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
      except Exception:
          print('###error.i:', i, line)
    return examples

# This function is not used by this file but is still used by the Colab and
# people who depend on it.
def convert_examples_to_features(examples, label_list, max_seq_length,
                                 tokenizer):
  """Convert a set of `InputExample`s to a list of `InputFeatures`."""

  features = []
  for (ex_index, example) in enumerate(examples):
    if ex_index % 10000 == 0:
      tf.logging.info("Writing example %d of %d" % (ex_index, len(examples)))

    feature = convert_single_example(ex_index, example, label_list,
                                     max_seq_length, tokenizer)

    features.append(feature)
  return features



#
from src.rank.rank_base import RankBase
import numpy as np

class BertRank(RankBase):
    def predict(self, scores, ids, num):
        if num == 1:
            scores = np.array(scores)
            max_id = np.argmax(scores)
            return [(scores[max_id], ids[max_id])]
        else:
            zipres = sorted([x for x in zip(scores, ids)],key=lambda x : x[0], reverse=True)[:num]
            return zipres




#!/anaconda3/bin/python
#coding=utf8
'''
# Author: Gavin King
# Created Time : 二 10/29 17:15:55 2019
'''

import yaml
import sys
import time
import os
rootpath=os.path.abspath(os.path.dirname(__file__))[:-3]
sys.path.append(rootpath)

from src.analysis.analysis_strategy import AnalysisStrategy
from src.retrieval.retrieval_strategy import RetrievalStrategy
from src.matching.matching_strategy import MatchingStrategy
from src.rank.rank_strategy import RankStrategy
from src.common.utils import loaddata, loaddict, QueryInfo, ResultInfo
import logging
from src.common.logger import setlogger
logger = logging.getLogger("QAsearch")

class QAsearch(object):
    def __init__(self, bertip=None, bertport=None):
        # 暂不支持conf地址修改
        conf_path = "conf/simqp.yaml"
        self.status = "fail"
        pwdpath = os.getcwd()
        os.chdir(rootpath)
        try:
            self.config = yaml.safe_load(open(conf_path, 'r', encoding='utf-8'))
            if bertip != None and bertport != None:
                self.config['matching_config']['bert']['server_ip'] = bertip
                self.config['matching_config']['bert']['server_port'] = bertport
            dict_map = loaddict(self.config['dict_config'])
            # 各个子模块的初始化
            self._analysis = AnalysisStrategy(dict_map, self.config['analysis_config'])
            self._retrieval = RetrievalStrategy(dict_map, self.config['retrieval_config'])
            self._matching = MatchingStrategy(dict_map, self.config['matching_config'])
            self._rank = RankStrategy(dict_map, self.config['rank_config'])
            self.dict_map = dict_map
        except:
            os.chdir(pwdpath)
            logger.error("QAsearch init failed")
            raise
        os.chdir(pwdpath)
        setlogger(self.config['logger'])

    def build(self, copus_path=rootpath+"data/corpus-chatbot.txt"):
        idx2ques, idx2label = loaddata(copus_path)
        corpus = self._analysis.buildCorpus(idx2ques, idx2label)
        self.dict_map['corpus'] = corpus
        self._retrieval.buildIndex(corpus)
        self.status = "succ"

    def add_corpus(self, corpus_path=None, corpus=None):
        if corpus_path:
            idx2ques, idx2label = loaddata(corpus_path)
        else:
            idx2ques, idx2label = corpus
        corpus = self._analysis.buildCorpus(idx2ques, idx2label)
        self.dict_map['corpus'] += corpus
        self._retrieval.add_index(corpus)


    def rebuild(self):
        pass

    def query(self, text, num=1):
        if self.status != "succ":
            return None
        try:
            qinfo = QueryInfo(text)
            result = ResultInfo()
            t1 = time.time()
            self._analysis.process(qinfo)
            logger.debug("analysis cost:{}".format(time.time()-t1))
            t1 = time.time()
            self._retrieval.retrieve(qinfo, result)
            logger.debug("retrieval cost:{}".format(time.time()-t1))
            t1 = time.time()
            self._matching.match(qinfo, result)
            logger.debug("matching cost:{}".format(time.time()-t1))
            t1 = time.time()
            self._rank.rank(qinfo, result, num)
            logger.info("rank cost:{}".format(time.time()-t1))
            logger.debug(" ".join([str(i) for i in result.retrieval['allids']]))
            logger.debug(" ".join([str(i) for i in result.matching['bert']]))
            logger.debug(" ".join([str(i[0])+"-"+str(i[1]) for i in result.rank['final']]))
            if num == 1:
                res = (self.dict_map['corpus'][0][int(result.rank['final'][0][1])], self.dict_map['corpus'][1][int(result.rank['final'][0][1])], result.rank['final'][0][0])
            else:
                res = [(self.dict_map['corpus'][0][int(result.rank['final'][i][1])], self.dict_map['corpus'][1][int(result.rank['final'][i][1])], result.rank['final'][i][0]) for i in range(num)]
            return res
        except Exception as e:
            logger.error(e)
            return "","",0.

if __name__ == "__main__":
    #simqp_conf_path = r'./conf/simqp.yaml'
    #copus_path = r'./data/samples.txt'
    #copus_path = "./data/corpus-chatbot.txt"

    Simqp_ins = QAsearch()
    Simqp_ins.build()
    query = "你是谁"
    print(query)
    result = Simqp_ins.query(query, num=3)
    print(result)
    query = "你喜欢什么"
    print(query)
    result = Simqp_ins.query(query)
    print(result)



from src.common.utils import load_func

class RetrievalStrategy:
    def __init__(self, dict_map, retireval_config):
        self.dict_map = dict_map
        # 读取配置文件
        self.retireval_config = retireval_config
        # 用于存放所有的方法实例
        self.method_dict = {}
        if len(self.retireval_config) < 1:
            raise ValueError("len(Retrieval plugin)==0")
        for name, method_config in self.retireval_config.items():
            if method_config['use']:
                func, other_params = load_func(method_config)
                retrieval_tool = func(other_params)
                self.method_dict[name] = retrieval_tool

    def buildIndex(self, corpus):
        for name in self.method_dict:
            self.method_dict[name]._build_index(corpus)

    def add_index(self, extra_corpus):
        for name in self.method_dict:
            if name.lower() == 'faiss':
                self.method_dict[name].add_indenx(extra_corpus)

    def merge(self, resultinfo):
        pass

    def retrieve(self, queryinfo, resultinfo):
        '''
        返回的是召回的id集合， matching 部分需要通过id集合从全局字典中取值。
        '''
        # 各分支分别召回, 返回idx的列表
        for name, retireval_tool in self.method_dict.items():
                retireval_tool.search(queryinfo, resultinfo)
        if len(self.method_dict) > 1:
            self.merge(resultinfo)



import numpy as np
import os

from gensim.models import KeyedVectors
from src.analysis.analysis_base import AnalysisBase
import logging
from src.QAsearch import rootpath
logger = logging.getLogger("QAsearch")

class SenEmbeddingBase(AnalysisBase):

    def __init__(self, vec_path=None, remote_path=None):
        self.vec_path = vec_path
        self.remote_path = remote_path
        self._load_vec()
        self.word2id = dict(
            zip(self.w2v.index2word, range(len(self.w2v.index2word))))

    def _load_vec(self):
        if self.vec_path is not None and os.path.exists(self.vec_path):
            if self.vec_path.endswith('model'):
                self.w2v = KeyedVectors.load(self.vec_path)
            elif self.vec_path.endswith('txt'):
                self.w2v = KeyedVectors.load_word2vec_format(
                    self.vec_path, binary=False)
            else:
                raise FileNotFoundError(
                    f"The embedding file: {self.vec_path} is not supported to parse ! ")
        elif self.remote_path is not None:
            print("the first time, need to download the emb file ,please wait 1min....")
            ret = os.system("hadoop fs -get "+self.remote_path+" "+ self.vec_path)
            if ret != 0:
                os.system("cp "+self.vec_path+".small "+self.vec_path)
                print("ATTENTION!!!! can't connect to hdfs, use the partial embfile now, please download the full embfile manually to [{}]/data/embedding.txt!!!!".format(rootpath))
            if self.vec_path.endswith('model'):
                self.w2v = KeyedVectors.load(self.vec_path)
            elif self.vec_path.endswith('txt'):
                self.w2v = KeyedVectors.load_word2vec_format(
                    self.vec_path, binary=False)
            else:
                raise FileNotFoundError(
                    f"The embedding file: {self.vec_path} is not supported to parse ! ")
        else:
            raise FileNotFoundError("The w2v file is not exists ! ")
        logger.info("load vec successfully")

    def get_sen_emb(self, line):
        # line repr one unit, maybe segs[list] or text[str] or others
        raise NotImplementedError(
            "SenEmbedding mush have the function about get_sen_emb")

    def process(self, line):
        return self.get_sen_emb(line)

class SenEmbeddingSSC(SenEmbeddingBase):
    def __init__(self, dict_map, config):
        self.dict_map = dict_map
        self.config = config
        vec_path = ""
        remote_path = ""
        if 'vec_path' in config:
            vec_path = config['vec_path']
        if 'remote_path' in config:
            remote_path = config['remote_path']

        SenEmbeddingBase.__init__(self, vec_path, remote_path)
        # tokens must exists in SenEmbeddingSSC
        #if 'segment_token_basic' not in config['use_dict_name']:
        #    raise KeyError("SenEmbeddingSSC not have segment_token_basic")
        if 'sen_vec_size' not in config:
            raise KeyError("SenEmbeddingSSC not have sen_vec_size")
        self.sen_vec_size = config["sen_vec_size"]
        self._cache_vec = {}

    def _new_vector(self, v):
        return v if v is not None else np.random.randn(self.w2v.vector_size).astype("float32")

    def _max_match_segment(self, text, window_max=5):
        match = False
        idx = 0
        word = []
        while idx < len(text):
            for i in range(window_max, 0, -1):
                cand = text[idx:idx + i]
                if cand in self.w2v:
                    word.append(cand)
                    match = True
                    break
            if not match:
                i = 1
                word.append(text[idx])
            idx += 1
        return word

    def _get_word_emb(self, word):
        v = None
        if word is None:
            return None
        if word in self._cache_vec:
            return self._cache_vec[word]
        if word in self.w2v:
            v = self.w2v[word]
        else:
            words = self._max_match_segment(word)
            vector = np.zeros(self.w2v.vector_size, dtype="float32")
            for token in words:
                if token in self.w2v:
                    vector += self.w2v[token]
                else:
                    # ignore word not in vocab after the max_match_segment function
                    vector += np.zeros(self.w2v.vector_size).astype("float32")
            v = vector / len(words)
        v = self._new_vector(v)
        self._cache_vec[word] = v
        return v

    def get_sen_emb(self, tokens_basic):
        '''
        输入参数，格式如：(terms, tf, idf_dict)
        '''
        # terms repr sentence after tokenizer; list
        terms, tf, idf_dict = tokens_basic
        terms_id = [self.word2id[w] for w in terms if w in self.word2id]
        if len(terms) < 1:
            return (np.zeros(self.sen_vec_size, dtype="float32"), terms_id)
        # build vector
        vector = np.zeros(self.sen_vec_size, dtype="float32")
        w = 0
        for term in terms:
            vector += self._get_word_emb(term) * idf_dict[term]
            w += idf_dict[term]
        if w == 0:
            return (vector, None)
        else:
            return (vector / w, terms_id)



from src.common.utils import load_func

class RankStrategy:
    def __init__(self, dict_map, rank_config):
        self.dict_map = dict_map
        self.rank_config = rank_config
        self.methods = {}
        for rank_name in self.rank_config:
            if not self.rank_config[rank_name]["use"]:
                continue
            func, _ = load_func(self.rank_config[rank_name])
            self.methods[rank_name] = func()

    def _merge(self):
        pass

    def rank(self, queryinfo, resultinfo, num):
        resultinfo.rank = {}
        for method in self.methods:
            resultinfo.rank[method] = self.methods[method].predict(resultinfo.matching[method], resultinfo.retrieval['allids'], num)
            resultinfo.rank['final'] = resultinfo.rank[method]
        if len(self.methods) > 0:
            self._merge()




"""
gRPC and REST client 
pre-process text and get predictions from models
"""
import sys
import json
import requests
from src.matching.semantic.bert.similarity import SimProcessor
from src.matching.matching_base import MatchingBase
import tensorflow as tf
tf.logging.set_verbosity(tf.logging.ERROR)

class BertMatcher(MatchingBase):
    def __init__(self, *args, **kwargs):
        super(BertMatcher, self).__init__(*args, **kwargs)
        self.model_name = 'matching'
        self.model_version = self.config['model_version']
        self.max_seq_length = 512
        self.ip = self.config['server_ip']
        self.rest_port = self.config['server_port']

        # REST
        self.model_base_url = 'http://{}:{}/v1/models/matching/versions/{}:predict'.format(self.ip, self.rest_port, self.model_version)
        self.max_seq_length = self.config['max_seq_length']
        # init
        self.vocab = self.config['vocab_file']
        self.dp = SimProcessor(self.vocab, self.max_seq_length)

    def predict(self, texts):
        
        features = self.dp.generate_from_input(texts)
        return self.rest_client(features)

    def rest_client(self, features):
        headers = {"content-type": "application/json"}
        model_url = self.model_base_url
        features = json.dumps({"signature_name": "serving_default", "instances": features})
        json_response = requests.post(model_url, data=features, headers=headers)
        if json_response.status_code == 200:
            predictions = json.loads(json_response.text)['predictions']
        else:
            predictions = None
        return predictions

    def compute_similarity(self, query_info, retrived_indices):
        
        query = query_info.textclean
        tuples = [(query, self.dict_map['corpus'][0][s]) for s in retrived_indices]
        
        predictions = self.predict(tuples)
        preds = [x[1] for x in predictions]
        return preds




import sys
import types
from src.matching.semantic.bert_matcher import BertMatcher
from src.matching.semantic.embedding_matcher import EmbeddingCosineMatcher
from src.matching.lexical.bm25_matcher import BM25Matcher
from src.matching.lexical.jaccard_matcher import CharJaccardMatcher, WordJaccardMatcher
from src.matching.lexical.tfidf_matcher import TfidfCosineMatcher
def str_to_class(classname):
    try:
        return getattr(sys.modules[__name__], classname)
    except:
        raise NotImplementedError


class MatchingStrategy:
    def __init__(self, dict_map, matching_config):
        self.dict_map = dict_map
        self.matching_config = matching_config
        models = list(self.matching_config.keys())
        self.methods = {}
        for model in models:
            class_nm = self.matching_config[model]['type']
            if self.matching_config[model]['use']:
                self.methods[model] = str_to_class(class_nm)(self.dict_map, self.matching_config[model])

    def match(self, queryinfo, resultinfo):
        ids = resultinfo.retrieval['allids']
        for method in self.methods:
            resultinfo.matching[method] = self.methods[method].compute_similarity(queryinfo, ids)
        return resultinfo
        




import codecs
import os
import re
# import string
import jieba
import jieba.analyse
from collections import Counter, defaultdict
from src.analysis.analysis_base import AnalysisBase

class SegmentJiebaSSC(AnalysisBase):
    """
        Description:
            tokenize str into a list of word
    """

    def __init__(self, dict_map, config):
        self.dict_map = dict_map
        self.config = config
        self._load_userdict()
        self.stopwords = []
        if 'stopwords' in self.config['use_dict_name']:
            self.stopwords = self.dict_map['stopwords']
        # 去掉字母和数字
        self.prefilter = self.config['prefilter']
        # 是否移除停用词
        self.ifremove = self.config['ifremove']
        self.iflower = self.config['iflower']

    def _load_userdict(self):
        self.ents = []
        if 'user_dict_path' in self.config and os.path.exists(self.config['user_dict_path']):
            jieba.load_userdict(self.config['user_dict_path'])
            self.ents = self.dict_map['user_dict_path']

    def extract_tags(self, text):
        return jieba.analyse.extract_tags(text, topK=10, withWeight=True)

    def seg(self, text):
        tokens = []
        if self.iflower and isinstance(text, str):
            text = text.lower()
        # python3
        if self.prefilter:  # 去掉所有字母和数字
            text = re.sub(r"[\w\d]+", " ", text, flags=re.ASCII)

        for x in jieba.cut(text, cut_all=False):
            x = x.strip()
            if self.ifremove and x in self.stopwords:
                continue
            if len(x) < 1:
                continue
            tokens.append(x)
        return tokens

    def process(self, text):
        terms = self.seg(text)
        tf = Counter(terms)
        idf_dict = defaultdict(float)
        for i, v in self.extract_tags(text):
            if i in self.ents:
                if v < 5:
                    idf_dict[i] = v + 1
                else:
                    idf_dict[i] = v
            else:
                idf_dict[i] = v
        for term in terms:
            if not term in idf_dict:
                if term in self.ents:
                    idf_dict[term] = 5
                else:
                    idf_dict[term] = 0.5
        return (terms, tf, idf_dict)


class SegmentHanlp(AnalysisBase):

    def __init__(self, dict_map, config):
        from pyhanlp import HanLP
        self.dict_map = dict_map
        self.config = config
        self.stopwords = []
        if 'stopwords' in self.config['use_dict_name']:
            self.stopwords = self.dict_map['stopwords']
        # 去掉字母和数字
        self.prefilter = False
        # 是否移除停用词
        self.ifremove = True
        self.iflower = True

    def seg(self, text):
        tokens = []
        if self.iflower:
            text = text.lower()
        # python3
        if self.prefilter:  # 去掉所有字母和数字
            text = re.sub(r"[\w\d]+", " ", text, flags=re.ASCII)
        for x in HanLP.segment(text, cut_all=False):
            x = x.word.strip()
            if self.ifremove and x in self.stopwords:
                continue
            if len(x) < 1:
                continue
            tokens.append(x)
        return tokens

    def process(self, text):
        return self.seg(text)



import importlib
import os


def load_func(config):
    bath_path = config['name']
    bath_type = config['type']
    other_parms = {k:config[k] for k in config if k not in ["name", "type"]} 
    print(bath_path)
    m = importlib.import_module(bath_path)
    func = getattr(m, bath_type)
    return func, other_parms

def loaddata(path):
    idx2ques = {}
    idx2label = {}
    with open(path, 'r', encoding='utf-8') as f:
        for line in f:
            items = line.strip().split("\t")
            idx2ques[int(items[-1])] = items[0]
            idx2label[int(items[-1])] = items[1]
    print("load data done")
    return idx2ques, idx2label

def loaddict(dict_config):
    all_dict = {}
    for config in dict_config:
        dict_name = config['name']
        dict_path = config['path']
        if not os.path.exists(dict_path):
            raise FileExistsError(f"dict load failed, {dict_path} not exist")
        with open(dict_path, 'r', encoding='utf-8') as f:
            all_dict[dict_name] = [line.strip() for line in f.readlines()]
    return all_dict

class QueryInfo(object):
    def __init__(self, text):
        self.text = text
        self.textclean = None
        self.terminfo = None
        self.embinfo = None

class ResultInfo(object):
    def __init__(self):
        self.retrieval = {}
        self.matching = {}
        self.ranking = {}



from src.matching.matching_base import MatchingBase
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

class EmbeddingCosineMatcher(MatchingBase):
    def computer_simlarity(self, query_info, retrived_indices):
        query = query_info.embinfo[0]
        samples = [self.dict_map['corpus'][3][sample] for sample in retrived_indices]
        return self._calc_cos(samples, query)

    def _calc_cos(self, docs, query):
        cosineSimilarities = cosine_similarity(query, docs).flatten()
        return cosineSimilarities




from setuptools import setup, find_packages
import io

setup(name='qatools-ifchange',\
        version='0.0.7',\
        description='fix encoding version',\
        #long_description=open('README.md',encoding='utf-8').read(),\
        #long_description=io.open('./QAplatform/README.md','r',-1,'utf-8').read(),\
        long_description_content_type="text/markdown",\
        url='https://github.com',\
        author='LH19880520',\
        author_email='peng.wang@ifchange.com',\
        license='ifchange',\
        install_requires=['tensorflow>=1.12.0','pyyaml','requests','sklearn','rank_bm25','jieba','gensim','faiss-cpu','tornado==5.0.2'],\
        packages=find_packages(),\
        package_data={'QAplatform': ['conf/*','data/*']},\
        include_package_data=True,\
        zip_safe=False)



class MatchingBase:
    def __init__(self, dict_map, config):
        self.dict_map = dict_map
        self.config = config

    def compute_simlarity(self, query_info, retrived_indices):
        raise NotImplementedError("Matching plugin must have computer_similartiy function")




#
from src.common.utils import load_func
class AnalysisStrategy:
    def __init__(self, dict_map, analysis_config):
        self.dict_map = dict_map
        self.analysis_config = analysis_config
        self.bather = None
        self.tokenizer = None
        self.embedder = None
        if 'dataclean' in self.analysis_config and self.analysis_config['dataclean']['use']:
            dataclean_config = self.analysis_config['dataclean']
            func, other_parms = load_func(dataclean_config)
            self.bather = func(self.dict_map, other_parms)
        if 'segment' in self.analysis_config and self.analysis_config['segment']['use']:
            segment_config = self.analysis_config['segment']
            func, other_parms = load_func(segment_config) 
            self.tokenizer = func(self.dict_map, other_parms)
        if 'embedding' in self.analysis_config and self.analysis_config['embedding']['use']:
            embedding_config = self.analysis_config['embedding']
            func, other_parms = load_func(embedding_config) 
            self.embedder = func(self.dict_map, other_parms)

    def buildCorpus(self, idx2ques, idx2label):
        corpus = {}
        idx2ques_bath = idx2ques
        if self.bather:
            idx2ques_bath = self.bather.value_process(idx2ques)
        # 翻转去重,洗完澡后可能会重复
        ques2idx = {ques: idx for idx, ques in idx2ques_bath.items()}
        idx2ques = {idx: ques for ques, idx in ques2idx.items()}
        idx2label = {idx: label for idx, label in idx2label.items() if idx in idx2ques}

        idx2term = None
        idx2emb = None
        if self.tokenizer:
            idx2term = self.tokenizer.value_process(idx2ques)

        if self.tokenizer and self.embedder:
            idx2emb = self.embedder.value_process(idx2term)

        return (idx2ques, idx2label, idx2term, idx2emb)
    
    def process(self, queryinfo):
        # 对输入的query进行分析
        text = queryinfo.text
        textclean = text
        terminfo = None
        embinfo = None
        if self.bather:
            textclean = self.bather.process(text)
        if self.tokenizer:
            # (terms, tf, idf_dict)
            terminfo = self.tokenizer.process(textclean)
        if self.embedder and self.tokenizer:
            # (vec, terms_id)
            embinfo = self.embedder.process(terminfo)
        queryinfo.textclean = textclean
        queryinfo.terminfo = terminfo
        queryinfo.embinfo = embinfo



"""
进行文本相似度预测的示例。可以直接运行进行预测。
参考了项目：https://github.com/chdd/bert-utils

"""

from .run_classifier import InputExample, DataProcessor, convert_examples_to_features
import src.matching.semantic.bert.tokenization as tokenization


# os.environ['CUDA_VISIBLE_DEVICES'] = '1'


class SimProcessor(DataProcessor):

    def __init__(self, vocab, max_sent_len):
        self.vocab = vocab
        self.max_sent_len = max_sent_len
        self.tokenizer = tokenization.FullTokenizer(vocab_file=self.vocab, do_lower_case=True)

    def get_sentence_examples(self, questions):
        examples = []
        for index, data in enumerate(questions):
            guid = 'test-%d' % index
            text_a = tokenization.convert_to_unicode(str(data[0]))
            text_b = tokenization.convert_to_unicode(str(data[1]))
            label = str(0)
            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples

    def get_labels(self):
        return ['0', '1']

    def generate_from_input(self, sentences):
        #tokenizer = tokenization.FullTokenizer(vocab_file=self.vocab, do_lower_case=True)

        predict_examples = self.get_sentence_examples(sentences)
        features = convert_examples_to_features(predict_examples, self.get_labels(), self.max_sent_len,
                                                self.tokenizer)
        res = []
        for feature in features:

            res += [{
                'input_ids': feature.input_ids,
                'input_mask': feature.input_mask,
                'segment_ids': feature.segment_ids,
                'label_ids': 0
            }]
        return res
    #
# if __name__ == '__main__':
#     sim = BertSim()
#     sim.start_model()
#     sim.predict_sentences([("我喜欢妈妈做的汤", "妈妈做的汤我很喜欢喝")])



class RetrievalBase:
    def build_index(self, *args, **kwargs):
        # 建立初始索引
        raise NotImplementedError("Retireval plugin must have build_index function")
    
    def search(self, qinfo, rinfo):
        # 搜索与当前值最相近的top_k的索引值
        raise NotImplementedError("Retrieval plugin must have search function")
